import json
import os
import re

from cumulusci.core.config import ServiceConfig
from cumulusci.core.exceptions import ServiceNotConfigured, TaskOptionsError
from cumulusci.core.keychain import DEFAULT_CONNECTED_APP
from cumulusci.core.utils import process_bool_arg
from cumulusci.oauth.client import PROD_LOGIN_URL
from cumulusci.tasks.sfdx import SFDX_CLI, SFDXBaseTask
from cumulusci.utils import random_alphanumeric_underscore, temporary_dir

CONNECTED_APP = """<?xml version="1.0" encoding="UTF-8"?>
<ConnectedApp xmlns="http://soap.sforce.com/2006/04/metadata">
    <contactEmail>{email}</contactEmail>
    <label>{label}</label>
    <oauthConfig>
        <callbackUrl>http://localhost:8080/callback</callbackUrl>
        <consumerKey>{client_id}</consumerKey>
        <consumerSecret>{client_secret}</consumerSecret>
        <scopes>Web</scopes>
        <scopes>Full</scopes>
        <scopes>RefreshToken</scopes>
    </oauthConfig>
</ConnectedApp>"""

PACKAGE_XML = """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <members>*</members>
        <name>ConnectedApp</name>
    </types>
    <version>44.0</version>
</Package>"""


class CreateConnectedApp(SFDXBaseTask):
    client_id_length = 85
    client_secret_length = 32
    deploy_wait = 5
    task_options = {
        "label": {
            "description": "The label for the connected app.  Must contain only alphanumeric and underscores",
            "required": True,
        },
        "email": {
            "description": "The email address to associate with the connected app.  Defaults to email address from the github service if configured."
        },
        "username": {
            "description": "Create the connected app in a different org.  Defaults to the target-dev-hub configured in sfdx.",
            "required": False,
        },
        "connect": {
            "description": "If True, the created connected app will be stored as the CumulusCI connected_app service in the keychain.",
            "required": False,
        },
        "overwrite": {
            "description": "If True, any existing connected_app service in the CumulusCI keychain will be overwritten.  Has no effect if the connect option is False.",
            "required": False,
        },
    }

    def _init_options(self, kwargs):
        self.client_id = None
        self.client_secret = None
        kwargs["command"] = "project deploy start --wait {}".format(self.deploy_wait)
        super(CreateConnectedApp, self)._init_options(kwargs)

        # Validate label
        if not re.match(r"^\w+$", self.options["label"]):
            raise TaskOptionsError(
                "label value must contain only alphanumeric or underscore characters"
            )

        # Default email to the github service's email if configured
        if "email" not in self.options:
            try:
                github = self.project_config.keychain.get_service("github")
            except ServiceNotConfigured:
                raise TaskOptionsError(
                    "Could not determine a default for option 'email'.  Either configure the github service using 'cci service connect github' or provide a value for the 'email' option"
                )
            self.options["email"] = github.email

        self.options["connect"] = process_bool_arg(self.options.get("connect") or False)
        self.options["overwrite"] = process_bool_arg(
            self.options.get("overwrite") or False
        )

    def _set_default_username(self):
        self.logger.info("Getting username for the default devhub from sfdx")
        output = []
        self._run_command(
            command="{} force config get target-dev-hub --json".format(SFDX_CLI),
            env=self._get_env(),
            output_handler=output.append,
        )
        self._process_devhub_output(b"\n".join(output))

    def _process_json_output(self, output):
        try:
            data = json.loads(output)
            return data
        except Exception:
            self.logger.error("Failed to parse json from output: {}".format(output))
            raise

    def _process_devhub_output(self, output):
        data = self._process_json_output(output)
        if "value" not in data["result"][0]:
            raise TaskOptionsError(
                "No sfdx config found for target-dev-hub.  Please use the sf force config set to set the target-dev-hub and run again"
            )
        self.options["username"] = data["result"][0]["value"]

    def _generate_id_and_secret(self):
        self.client_id = random_alphanumeric_underscore(self.client_id_length)
        self.client_secret = random_alphanumeric_underscore(self.client_secret_length)

    def _build_package(self):
        connected_app_path = "connectedApps"
        os.mkdir(connected_app_path)
        self._generate_id_and_secret()
        with open(
            os.path.join(connected_app_path, self.options["label"] + ".connectedApp"),
            "w",
        ) as f:
            f.write(
                CONNECTED_APP.format(
                    label=self.options["label"],
                    email=self.options["email"],
                    client_id=self.client_id,
                    client_secret=self.client_secret,
                )
            )
        with open("package.xml", "w") as f:
            f.write(PACKAGE_XML)

    def _validate_connect_service(self):
        if not self.options["overwrite"]:
            try:
                connected_app = self.project_config.keychain.get_service(
                    "connected_app", self.options["label"]
                )
            except ServiceNotConfigured:  # pragma: no cover
                pass
            else:
                if connected_app is not DEFAULT_CONNECTED_APP:
                    raise TaskOptionsError(
                        "The CumulusCI keychain already contains a connected_app service.  Set the 'overwrite' option to True to overwrite the existing service"
                    )

    def _connect_service(self):
        self.project_config.keychain.set_service(
            "connected_app",
            self.options["label"],
            ServiceConfig(
                {
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                    "login_url": PROD_LOGIN_URL,
                    "callback_url": "http://localhost:8080/callback",
                }
            ),
        )

    def _get_command(self):
        command = super()._get_command()
        # Default to sf target-dev-hub
        if "username" not in self.options:
            self._set_default_username()
        command += " -o {}".format(self.options.get("username"))
        command += " --metadata-dir {}".format(self.tempdir)
        return command

    def _run_task(self):
        if self.options["connect"]:
            self._validate_connect_service()

        with temporary_dir() as tempdir:
            self.tempdir = tempdir
            self._build_package()
            super()._run_task()

        if self.options["connect"]:
            self._connect_service()



from typing import List

from cumulusci.cli.runtime import CliRuntime
from cumulusci.cli.utils import group_items
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.tasks import BaseTask


class RetrieveTasks(BaseTask):
    task_options = {
        "group_name": {
            "description": "Name of the category or Group",
            "required": True,
        },
    }

    def _run_task(self):
        runtime = CliRuntime(load_keychain=True)
        tasks = runtime.get_available_tasks()
        task_groups = group_items(tasks)
        task_groups = task_groups[self.options["group_name"]]
        self.return_values: List[str] = []
        for task_name, description in task_groups:
            self.return_values.append(task_name)
        if self.return_values:
            self.return_values.sort()
        else:
            raise CumulusCIException("No tasks in the specified group")

        self.logger.info(self.return_values)



"""
A Template like this:

template.tpl:

<?xml ...><foo xmls=...><id>{id}</id><status>{status:Closed}</status></foo>

can be referred to by name like this:

include_file: template.tpl
vars:
    id: 001A2431AF3

Note that the status can be elided because it defaults to Closed.
"""


import re
import typing as T
from pathlib import Path

MATCHER_REGEXP = re.compile("{(?P<name>[^:}]+)(:(?P<default>.*))?}")


class Pattern(T.NamedTuple):
    """Template pattern that can be referred to in to save space"""

    filename: str  # where to find the template
    orig_pattern: str  # the contents of the file
    regexp: re.Pattern  # contents converted to Python compatible regexp
    replacement_template: str  # contents converted to Python format-string
    default_values: T.Dict[str, str]  # defaulted values to save space further

    @staticmethod
    def parse(filename, pat) -> "Pattern":
        """Parse the contents of a pattern file into a Pattern obj"""
        replaceables = list(MATCHER_REGEXP.findall(pat))
        assert replaceables
        escaped_pat = pat.replace("?", r"\?")  # re.escape is too aggressive :(
        regexp = re.compile(MATCHER_REGEXP.sub(r"(?P<\1>.*)", escaped_pat))
        replacement_template = MATCHER_REGEXP.sub(r"{\1}", pat)
        default_values = {match[0]: match[-1] for match in replaceables if match[-1]}
        return Pattern(filename, pat, regexp, replacement_template, default_values)


class StringToTemplateCompressor:
    """Code that can translate from strings to templates and back"""

    @classmethod
    def from_directory(cls, path: T.Union[Path, str]):
        """Make a compressor from a directory of "tpl" files."""
        path = Path(path)
        raw_patterns = {f.name: f.read_text() for f in path.glob("*.tpl")}
        assert raw_patterns, f"Directory was empty {path}"
        return cls(raw_patterns)

    def __init__(self, raw_patterns: T.Dict[str, str]):
        self.patterns_by_filename = {
            filename: Pattern.parse(filename, pattern)
            for filename, pattern in raw_patterns.items()
        }
        self.patterns = self.patterns_by_filename.values()

    def string_to_template_if_possible(self, string: str):
        """Return a template dict if a template is matching.

        Else return the original string."""
        match = _find_match(self.patterns, string)
        if match:
            pattern, values = match
            return {"include_template": pattern.filename, "vars": values}
        else:
            return string

    def template_to_string(self, template: dict):
        """Reverse the templating process"""
        template_name = template["include_template"]
        template_vars = template["vars"]
        template = self.patterns_by_filename[template_name]
        return template.replacement_template.format(
            **{**template.default_values, **template_vars}
        )


def _expand(pattern, values):
    return pattern.replacement_template.format(**{**pattern.default_values, **values})


def _find_match(patterns, string):
    for pattern in patterns:
        match = pattern.regexp.match(string)
        if match:
            return pattern, {
                k: v
                for k, v in match.groupdict().items()
                if v != pattern.default_values.get(k)
            }



import typing as T


class OrderedSet(dict):
    """Extremely minimal ordered set

    Use with care, because very few methods are implemented."""

    def __init__(self, other: T.Optional[T.Iterable] = None):
        if other is not None:
            self.update(other)

    def add(self, item):
        self[item] = item

    def remove(self, item):
        del self[item]

    def copy(self):
        return OrderedSet(self)

    def update(self, other: T.Iterable):
        super().update((x, x) for x in other)

    def union(self, other: T.Iterable):
        new = OrderedSet(self)
        new.update(other)
        return new


# Workaround for Python 3.8
OrderedSetType = T.Iterable



import logging

import pytest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.metadata_etl.field_sets import AddFieldsToFieldSet
from cumulusci.tasks.salesforce.tests.util import create_task
from cumulusci.utils.xml import metadata_tree

OBJECT_XML = b"""<?xml version="1.0" encoding="UTF-8"?>
<CustomObject xmlns="http://soap.sforce.com/2006/04/metadata">
    <fieldSets>
        <fullName>IncidentDetail</fullName>
        <availableFields>
            <field>IncidentDateTime__c</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </availableFields>
        <availableFields>
            <field>LocationNotes__c</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </availableFields>
        <availableFields>
            <field>WatchList__c</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </availableFields>
        <description>This field set is used by the Incidents component</description>
        <displayedFields>
            <field>ContactId</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </displayedFields>
        <displayedFields>
            <field>Subject</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </displayedFields>
        <displayedFields>
            <field>Description</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </displayedFields>
        <displayedFields>
            <field>IncidentType__c</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </displayedFields>
        <displayedFields>
            <field>CreatedDate</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </displayedFields>
        <displayedFields>
            <field>ClosedDate</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </displayedFields>
        <displayedFields>
            <field>Branch__c</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </displayedFields>
        <displayedFields>
            <field>Severity__c</field>
            <isFieldManaged>false</isFieldManaged>
            <isRequired>false</isRequired>
        </displayedFields>
        <label>Incident Detail</label>
    </fieldSets>
</CustomObject>
"""


class TestAddFieldsToFieldSet:
    def test_adds_fields(self):
        task = create_task(
            AddFieldsToFieldSet,
            {
                "field_set": "Case.IncidentDetail",
                "fields": ["IncidentDateTime__c", "LocationNotes__c"],
            },
        )

        # Validate that the fields have been added as displayedFields
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "Case")
        fields = result.find("fieldSets", fullName="IncidentDetail").findall(
            "displayedFields"
        )

        test_elems = [f for f in fields if f.field.text == "LocationNotes__c"]
        assert len(test_elems) == 1
        assert test_elems[0].isFieldManaged.text == "false"
        assert test_elems[0].isRequired.text == "false"

        test_elems = [f for f in fields if f.field.text == "LocationNotes__c"]
        assert len(test_elems) == 1
        assert test_elems[0].isFieldManaged.text == "false"
        assert test_elems[0].isRequired.text == "false"

        # Validate that the fields were removed from availableFields
        fields = result.find("fieldSets", fullName="IncidentDetail").findall(
            "availableFields"
        )
        assert "IncidentDateTime__c" not in (f.field.text for f in fields)
        assert "LocationNotes__c" not in (f.field.text for f in fields)

        # Validate that the other field remains in availableFields
        assert "WatchList__c" in (f.field.text for f in fields)

    def test_does_not_add_duplicate_values(self, caplog):
        task = create_task(
            AddFieldsToFieldSet,
            {
                "field_set": "Case.IncidentDetail",
                "fields": ["ContactId"],
            },
        )

        caplog.set_level(logging.INFO)

        # Validate that the duplicate entry is not added
        tree = metadata_tree.fromstring(OBJECT_XML)
        result = task._transform_entity(tree, "Case")
        fields = result.find("fieldSets", fullName="IncidentDetail").findall(
            "displayedFields"
        )

        test_elems = [f for f in fields if f.field.text == "ContactId"]
        assert len(test_elems) == 1

        # Validate that info was logged
        assert "ContactId" in caplog.text

    def test_init_options__no_fields(self):
        with pytest.raises(TaskOptionsError):
            create_task(
                AddFieldsToFieldSet,
                {
                    "field_set": "Case.IncidentDetail",
                },
            )

    def test_init_options__bad_picklist_name(self):
        task = create_task(
            AddFieldsToFieldSet,
            {
                "field_set": "Case.Garbage",
                "fields": ["IncidentDateTime__c", "LocationNotes__c"],
            },
        )
        tree = metadata_tree.fromstring(OBJECT_XML)
        with pytest.raises(TaskOptionsError):
            task._transform_entity(tree, "Case")



import json
from pathlib import Path

import click
from rich.console import Console
from rst2ansi import rst2ansi

from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import CumulusCIUsageError
from cumulusci.utils import doc_task

from .runtime import pass_runtime
from .ui import CliTable
from .utils import group_items


@click.group("task", help="Commands for finding and running tasks for a project")
def task():
    pass


# Commands for group: task


@task.command(name="list", help="List available tasks for the current context")
@click.option("--plain", is_flag=True, help="Print the table using plain ascii.")
@click.option("--json", "print_json", is_flag=True, help="Print a json string")
@pass_runtime(require_project=False)
def task_list(runtime, plain, print_json):
    tasks = runtime.get_available_tasks()
    plain = plain or runtime.universal_config.cli__plain_output

    console = Console()
    if print_json:
        console.print_json(json.dumps(tasks))
        return None

    task_groups = group_items(tasks)
    for group, tasks in task_groups.items():
        data = [["Task", "Description"]]
        data.extend(sorted(tasks))
        table = CliTable(
            data,
            group,
        )
        console.print(table)

    console.print(
        "Use [bold]cci task info <task_name>[/] to get more information about a task."
    )


@task.command(name="doc", help="Exports RST format documentation for all tasks")
@click.option(
    "--project", "project", is_flag=True, help="Include project-specific tasks only"
)
@click.option(
    "--write",
    "write",
    is_flag=True,
    help="If true, write output to a file (./docs/project_tasks.rst or ./docs/cumulusci_tasks.rst)",
)
@pass_runtime(require_project=False)
def task_doc(runtime, project=False, write=False):
    if project and runtime.project_config is None:
        raise click.UsageError(
            "The --project option can only be used inside a project."
        )
    if project:
        full_tasks = runtime.project_config.tasks
        selected_tasks = runtime.project_config.config_project.get("tasks", {})
        file_name = "project_tasks.rst"
        project_name = runtime.project_config.project__name
        title = f"{project_name} Tasks Reference"
    else:
        full_tasks = selected_tasks = runtime.universal_config.tasks
        file_name = "cumulusci_tasks.rst"
        title = "Tasks Reference"

    result = ["=" * len(title), title, "=" * len(title), ""]
    for name, task_config_dict in full_tasks.items():
        if name not in selected_tasks:
            continue
        task_config = TaskConfig(task_config_dict)
        doc = doc_task(name, task_config)
        result += [doc, ""]
    result = "\n".join(result)

    if write:
        Path("docs").mkdir(exist_ok=True)
        (Path("docs") / file_name).write_text(result, encoding="utf-8")
    else:
        click.echo(result)


@task.command(name="info", help="Displays information for a task")
@click.argument("task_name")
@pass_runtime(require_project=False, require_keychain=True)
def task_info(runtime, task_name):
    task_config = (
        runtime.project_config.get_task(task_name)
        if runtime.project_config is not None
        else runtime.universal_config.get_task(task_name)
    )

    doc = doc_task(task_name, task_config).encode()
    click.echo(rst2ansi(doc))


class RunTaskCommand(click.MultiCommand):
    # options that are not task specific
    global_options = {
        "no-prompt": {
            "help": "Disables all prompts. Set for non-interactive mode such as calling from scripts or CI sytems",
            "is_flag": True,
        },
        "debug": {
            "help": "Drops into the Python debugger on an exception",
            "is_flag": True,
        },
        "debug-before": {
            "help": "Drops into the Python debugger right before the task starts",
            "is_flag": True,
        },
        "debug-after": {
            "help": "Drops into the Python debugger at task completion.",
            "is_flag": True,
        },
    }

    def list_commands(self, ctx):
        runtime = ctx.obj
        tasks = runtime.get_available_tasks()
        return sorted([t["name"] for t in tasks])

    def get_command(self, ctx, task_name):
        runtime = ctx.obj
        if runtime.project_config is None:
            raise runtime.project_config_error
        runtime._load_keychain()
        task_config = runtime.project_config.get_task(task_name)

        if "options" not in task_config.config:
            task_config.config["options"] = {}

        task_class = task_config.get_class()
        task_options = task_class.task_options

        params = self._get_default_command_options(task_class.salesforce_task)
        params.extend(self._get_click_options_for_task(task_options))

        def run_task(*args, **kwargs):
            """Callback function that executes when the command fires."""
            org, org_config = runtime.get_org(
                kwargs.pop("org", None), fail_if_missing=False
            )

            # Merge old-style and new-style command line options
            old_options = kwargs.pop("o", ())
            new_options = {
                k: v for k, v in kwargs.items() if k not in self.global_options
            }
            options = self._collect_task_options(
                new_options, old_options, task_name, task_options
            )

            # Merge options from the command line into options from the task config.
            task_config.config["options"].update(options)

            try:
                task = task_class(
                    task_config.project_config, task_config, org_config=org_config
                )

                if kwargs.get("debug_before", None):
                    import pdb

                    pdb.set_trace()

                task()

                if kwargs.get("debug_after", None):
                    import pdb

                    pdb.set_trace()

            finally:
                runtime.alert(f"Task complete: {task_name}")

        cmd = click.Command(task_name, params=params, callback=run_task)
        cmd.help = task_config.description
        return cmd

    def format_help(self, ctx, formatter):
        """Custom help for `cci task run`"""
        runtime = ctx.obj
        tasks = runtime.get_available_tasks()
        task_groups = group_items(tasks)
        console = Console()
        for group, tasks in task_groups.items():
            data = [["Task", "Description"]]
            data.extend(sorted(tasks))
            table = CliTable(
                data,
                group,
            )
            console.print(table)

        console.print("Usage: cci task run <task_name> [TASK_OPTIONS...]\n")
        console.print("See above for a complete list of available tasks.")
        console.print(
            "Use [bold]cci task info <task_name>[/] to get more information about a task and its options."
        )

    def _collect_task_options(self, new_options, old_options, task_name, task_options):
        """Merge new style --options with old style -o options.

        Raises:
            CumulusCIUsageError: if there is an old option which duplicates a new one,
            or the option doesn't exist for the given task.
        """
        # filter out options with no values
        options = {
            normalize_option_name(k): v for k, v in new_options.items() if v is not None
        }

        for k, v in old_options:
            k = normalize_option_name(k)
            if options.get(k):
                raise CumulusCIUsageError(
                    f"Please make sure to specify options only once. Found duplicate option `{k}`."
                )
            if k not in task_options:
                raise CumulusCIUsageError(
                    f"No option `{k}` found in task {task_name}.\nTo view available task options run: `cci task info {task_name}`"
                )
            options[k] = v
        return options

    def _get_click_options_for_task(self, task_options):
        """
        Given a dict of options in a task, constructs and returns the
        corresponding list of click.Option instances
        """
        click_options = [click.Option(["-o"], nargs=2, multiple=True, hidden=True)]
        for name, properties in task_options.items():
            # NOTE: When task options aren't explicitly given via the command line
            # click complains that there are no values for options. We set required=False
            # to mitigate this error. Task option validation should be performed at the
            # task level via task._validate_options() or Pydantic models.
            decls = set(
                (
                    f"--{name}",
                    f"--{name.replace('_', '-')}",
                )
            )

            click_options.append(
                click.Option(
                    param_decls=tuple(decls),
                    required=False,  # don't enforce option values in Click
                    help=properties.get("description", ""),
                )
            )
        return click_options

    def _get_default_command_options(self, is_salesforce_task):
        click_options = []
        for opt_name, config in self.global_options.items():
            click_options.append(
                click.Option(
                    param_decls=(f"--{opt_name}",),
                    is_flag=config["is_flag"],
                    help=config["help"],
                )
            )

        if is_salesforce_task:
            click_options.append(
                click.Option(
                    param_decls=("--org",),
                    help="Specify the target org. By default, runs against the current default org.",
                )
            )

        return click_options


@task.command(cls=RunTaskCommand, name="run", help="Runs a task")
def task_run():
    pass  # pragma: no cover


def normalize_option_name(k):
    return k.replace("-", "_")



import copy

from cumulusci.core.github import markdown_link_to_pr
from cumulusci.core.utils import process_bool_arg
from cumulusci.tasks.base_source_control_task import BaseSourceControlTask
from cumulusci.tasks.release_notes.generator import BaseReleaseNotesGenerator
from cumulusci.utils.deprecation import warn_moved
from cumulusci.vcs.bootstrap import (
    get_pull_requests_with_base_branch,
    is_label_on_pull_request,
    is_pull_request_merged,
)
from cumulusci.vcs.models import AbstractPullRequest, AbstractRepo


class AllVcsReleaseNotes(BaseSourceControlTask):
    filename: str = "vcs_release_notes.html"

    task_options = {
        "repos": {
            "description": (
                "The list of owner, repo key pairs for which to generate release notes."
                + " Ex: 'owner': SalesforceFoundation 'repo': 'NPSP'"
            ),
            "required": True,
        },
    }

    def _run_task(self):
        table_of_contents = "<h1>Table of Contents</h1><ul>"
        body = ""
        for project in self.options["repos"]:
            if project["owner"] and project["repo"]:
                options = copy.deepcopy(self.options)
                options["repo_owner"] = project["owner"]
                options["repo_name"] = project["repo"]

                release = (
                    self.vcs_service.get_repository(options=options)
                    .latest_release()
                    .body
                )
                table_of_contents += (
                    f"""<li><a href="#{project['repo']}">{project['repo']}</a></li>"""
                )
                release_project_header = (
                    f"""<h1 id="{project['repo']}">{project['repo']}</h1>"""
                )
                release_html = self.vcs_service.markdown(
                    release,
                    mode="gfm",
                    context="{}/{}".format(project["owner"], project["repo"]),
                )
                body += f"{release_project_header}<hr>{release_html}<hr>"
        table_of_contents += "</ul><br><hr>"
        head = "<head><title>Release Notes</title></head>"
        body = f"<body>{table_of_contents}{body}</body>"
        result = f"<html>{head}{body}</html>"
        with open(self.filename, "w") as f:
            f.write(result)


class VcsReleaseNotes(BaseSourceControlTask):
    task_options = {
        "tag": {
            "description": (
                "The tag to generate release notes for." + " Ex: release/1.2"
            ),
            "required": True,
        },
        "last_tag": {
            "description": (
                "Override the last release tag. This is useful"
                + " to generate release notes if you skipped one or more"
                + " releases."
            )
        },
        "link_pr": {
            "description": (
                "If True, insert link to source pull request at" + " end of each line."
            )
        },
        "publish": {"description": "Publish to VCS release if True (default=False)"},
        "include_empty": {
            "description": "If True, include links to PRs that have no release notes (default=False)"
        },
        "version_id": {
            "description": "The package version id used by the InstallLinksParser to add install urls"
        },
        "trial_info": {
            "description": "If True, Includes trialforce template text for this product."
        },
        "sandbox_date": {
            "description": "The date of the sandbox release in ISO format (Will default to None)"
        },
        "production_date": {
            "description": "The date of the production release in ISO format (Will default to None)"
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.options["link_pr"] = process_bool_arg(self.options.get("link_pr", False))
        self.options["publish"] = process_bool_arg(self.options.get("publish", False))
        self.options["include_empty"] = process_bool_arg(
            self.options.get("include_empty", False)
        )
        self.options["trial_info"] = self.options.get("trial_info", False)
        self.options["sandbox_date"] = self.options.get("sandbox_date", None)
        self.options["production_date"] = self.options.get("production_date", None)

    def _run_task(self):
        release_notes: BaseReleaseNotesGenerator = (
            self.vcs_service.release_notes_generator(self.options)
        )
        self.logger.info("\n" + release_notes())


class ParentPullRequestNotes(BaseSourceControlTask):
    task_docs = """
    Aggregate change notes from child pull request(s) to a corresponding parent pull request.

    When given the branch_name option, this task will: (1) check if the base branch
    of the corresponding pull request starts with the feature branch prefix and if so (2) attempt
    to query for a pull request corresponding to this parent feature branch. (3) if a pull request
    isn't found, the task exits and no actions are taken.

    If the build_notes_label is present on the pull request, then all notes from the
    child pull request are aggregated into the parent pull request. if the build_notes_label
    is not detected on the parent pull request then a link to the child pull request
    is placed under the "Unaggregated Pull Requests" header.

    If you have a pull request on branch feature/myFeature that you would like to rebuild notes
    for use the branch_name and force options:
        cci task run vcs_parent_pr_notes --branch-name feature/myFeature --force True
    """
    UNAGGREGATED_PR_HEADER = "\r\n\r\n# Unaggregated Pull Requests"

    task_options = {
        "branch_name": {
            "description": "Name of branch to check for parent status, and if so, reaggregate change notes from child branches.",
            "required": True,
        },
        "build_notes_label": {
            "description": (
                "Name of the label that indicates that change notes on parent pull "
                "requests should be reaggregated when a child branch pull request is created."
            ),
            "required": True,
        },
        "force": {
            "description": "force rebuilding of change notes from child branches in the given branch.",
            "required": False,
        },
    }

    def _init_options(self, kwargs):
        super(ParentPullRequestNotes, self)._init_options(kwargs)
        self.options["branch_name"] = self.options.get("branch_name")
        self.options["build_notes_label"] = self.options.get("build_notes_label")
        self.options["force"] = self.options.get("force")

    def _setup_self(self):
        self.repo: AbstractRepo = self.get_repo()
        self.commit = self.repo.get_commit(self.project_config.repo_commit)
        self.branch_name = self.options.get("branch_name")
        self.force_rebuild_change_notes = process_bool_arg(
            self.options["force"] or False
        )
        self.generator: BaseReleaseNotesGenerator = (
            self.vcs_service.parent_pr_notes_generator(self.repo)
        )

    def _run_task(self):
        self._setup_self()

        if self.force_rebuild_change_notes:
            pull_request = self._get_parent_pull_request()
            if pull_request:
                self.generator.aggregate_child_change_notes(pull_request)

        elif self._has_parent_branch() and self._commit_is_merge():
            parent_pull_request = self._get_parent_pull_request()
            if parent_pull_request:
                if is_label_on_pull_request(
                    self.repo,
                    parent_pull_request,
                    self.options.get("build_notes_label"),
                ):
                    self.generator.aggregate_child_change_notes(parent_pull_request)
                else:
                    child_branch_name = self._get_child_branch_name_from_merge_commit()
                    if child_branch_name:
                        self._update_unaggregated_pr_header(
                            parent_pull_request, child_branch_name
                        )

    def _has_parent_branch(self):
        feature_prefix = self.project_config.project__git__prefix_feature
        return (
            self.branch_name.startswith(feature_prefix) and "__" not in self.branch_name
        )

    def _commit_is_merge(self):
        return len(self.commit.parents) > 1

    def _get_parent_pull_request(self):
        """Attempts to retrieve a pull request for the given branch."""
        requests = get_pull_requests_with_base_branch(
            self.repo, self.repo.default_branch, self.branch_name
        )
        if len(requests) > 0:
            return requests[0]
        else:
            self.logger.info(f"Pull request not found for branch {self.branch_name}.")

    def _get_child_branch_name_from_merge_commit(self):
        pull_requests: list[
            AbstractPullRequest
        ] = self.repo.get_pull_requests_by_commit(self.commit.sha)
        merged_prs = list(filter(is_pull_request_merged, pull_requests))

        child_branch_name = None
        if len(merged_prs) == 1:
            return merged_prs[0].head.ref

        else:
            self.logger.error(
                f"Received multiple pull requests, expected one, for commit sha: {self.commit.sha}"
            )

        return child_branch_name

    def _update_unaggregated_pr_header(
        self, pull_request_to_update, branch_name_to_add
    ):
        """Updates the 'Unaggregated Pull Requests' section header with a link
        to the new child branch pull request"""

        self._add_header(pull_request_to_update)

        pull_requests = get_pull_requests_with_base_branch(
            self.repo,
            branch_name_to_add.split("__")[0],
            branch_name_to_add,
            state="all",
        )

        if len(pull_requests) == 0:
            self.logger.info(f"No pull request for branch {branch_name_to_add} found.")
        elif len(pull_requests) > 1:
            self.logger.error(
                f"Expected one pull request, found {len(pull_requests)} for branch {branch_name_to_add}"
            )
        else:
            self._add_link_to_pr(pull_request_to_update, pull_requests[0])

    def _add_header(self, pull_request):
        """Appends the header to the pull_request.body if not already present"""
        pull_request.body = "" if pull_request.body is None else pull_request.body
        if self.UNAGGREGATED_PR_HEADER not in pull_request.body:
            pull_request.body += self.UNAGGREGATED_PR_HEADER

    def _add_link_to_pr(self, to_update, to_link):
        """Updates pull request to_update with a link to pull
        request to_link if one does not already exist."""
        body = to_update.body
        pull_request_link = markdown_link_to_pr(to_link)
        if pull_request_link not in body:
            body += "\r\n* " + pull_request_link
            to_update.update(body=body)


# Deprecated aliases for backwards compatibility
# These classes are deprecated and will be removed in future versions.
# Use the new AllVcsReleaseNotes and VcsReleaseNotes classes instead.


class AllGithubReleaseNotes(AllVcsReleaseNotes):
    """Deprecated: use cumulusci.tasks.release_notes.task.AllVcsReleaseNotes instead"""

    filename: str = "github_release_notes.html"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn_moved(
            "cumulusci.tasks.release_notes.task.AllVcsReleaseNotes",
            f"{__name__}.AllGithubReleaseNotes",
        )


class GithubReleaseNotes(VcsReleaseNotes):
    """Deprecated: use cumulusci.tasks.release_notes.task.VcsReleaseNotes instead"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn_moved(
            "cumulusci.tasks.release_notes.task.VcsReleaseNotes",
            f"{__name__}.GithubReleaseNotes",
        )



import os
from unittest import mock

from cumulusci.core.flowrunner import StepSpec
from cumulusci.tasks.salesforce import DeployBundles
from cumulusci.utils import temporary_dir

from .util import create_task


class TestDeployBundles:
    def test_run_task(self):
        with temporary_dir() as path:
            os.mkdir("src")
            with open(os.path.join(path, "file"), "w"):
                pass
            task = create_task(DeployBundles, {"path": path})
            task._get_api = mock.Mock()
            task()
            task._get_api.assert_called_once()

    def test_run_task__path_not_found(self):
        with temporary_dir() as path:
            pass
        task = create_task(DeployBundles, {"path": path})
        task._get_api = mock.Mock()
        task()
        task._get_api.assert_not_called()

    def test_freeze(self):
        self.maxDiff = None
        with temporary_dir() as path:
            os.mkdir(".git")
            os.makedirs("unpackaged/test")
            task = create_task(DeployBundles, {"path": path + "/unpackaged"})
            step = StepSpec(
                step_num=1,
                task_name="deploy_bundles",
                task_config=task.task_config,
                task_class=None,
                project_config=task.project_config,
            )

            steps = task.freeze(step)
            print(steps)
            assert [
                {
                    "is_required": True,
                    "kind": "metadata",
                    "name": "Deploy unpackaged/test",
                    "path": "deploy_bundles.test",
                    "source": None,
                    "step_num": "1.1",
                    "task_class": "cumulusci.tasks.salesforce.UpdateDependencies",
                    "task_config": {
                        "options": {
                            "dependencies": [
                                {
                                    "ref": task.project_config.repo_commit,
                                    "github": "https://github.com/TestOwner/TestRepo",
                                    "subfolder": "unpackaged/test",
                                    "collision_check": False,
                                    "namespace_inject": None,
                                }
                            ]
                        },
                        "checks": [],
                    },
                }
            ] == steps

    def test_freeze__bad_path(self):
        task = create_task(DeployBundles, {"path": "/bogus"})
        step = StepSpec(1, "deploy_bundles", task.task_config, None, None)
        steps = task.freeze(step)
        assert [] == steps



import json
import sys
import time
from unittest import mock

import pytest
import requests
import responses
from packaging import version

import cumulusci

from .. import utils


def test_get_installed_version():
    result = utils.get_installed_version()
    assert cumulusci.__version__ == str(result)


@responses.activate
def test_get_latest_final_version():
    responses.add(
        method="GET",
        url="https://pypi.org/pypi/cumulusci-plus/json",
        body=json.dumps(
            {
                "releases": {
                    "1.0b1": {},
                    "1.0": {},
                    "1.0.1.dev0": {},
                    "1.0.1": {},
                    "1.0.post1": {},
                }
            }
        ),
        status=200,
    )
    result = utils.get_latest_final_version()
    assert result.base_version == "1.0.1"


@mock.patch("cumulusci.cli.utils.get_installed_version")
@mock.patch("cumulusci.cli.utils.get_latest_final_version")
@mock.patch("cumulusci.cli.utils.click")
def test_check_latest_version(click, get_latest_final_version, get_installed_version):
    with utils.timestamp_file() as f:
        f.write(str(time.time() - 4000))
    get_latest_final_version.return_value = version.parse("2")
    get_installed_version.return_value = version.parse("1")

    utils.check_latest_version()
    if sys.version_info > utils.LOWEST_SUPPORTED_VERSION:
        click.echo.assert_called_once()
    else:
        click.echo.assert_called()


@mock.patch("cumulusci.cli.utils.get_latest_final_version")
@mock.patch("cumulusci.cli.utils.click")
def test_check_latest_version_request_error(click, get_latest_final_version):
    with utils.timestamp_file() as f:
        f.write(str(time.time() - 4000))
    get_latest_final_version.side_effect = requests.exceptions.RequestException()

    utils.check_latest_version()

    click.echo.assert_any_call("Error checking cci version:", err=True)


@pytest.mark.skipif(
    not sys.platform.startswith("win"), reason="Requires Windows Registry"
)
@mock.patch("winreg.QueryValueEx")
def test_win32_warning(query_value):
    query_value.return_value = (1, 1)

    is_enabled = utils.win32_long_paths_enabled()

    query_value.assert_called_once()
    assert "LongPathsEnabled" in query_value.call_args.args
    assert is_enabled


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Should not print on non-win32"
)
@mock.patch("cumulusci.cli.utils.win32_long_paths_enabled")
@mock.patch("rich.console.Console.print")
def test_no_longpath_warn_on_posix(console_print, is_enabled):
    utils.warn_if_no_long_paths()

    is_enabled.assert_not_called()
    console_print.assert_not_called()


@pytest.mark.skipif(
    not sys.platform.startswith("win"), reason="Requires Windows Registry"
)
@mock.patch("cumulusci.cli.utils.win32_long_paths_enabled")
@mock.patch("rich.console.Console.print")
def test_no_longpath_warn_on_win(console_print, is_enabled):
    is_enabled.return_value = True

    utils.warn_if_no_long_paths()

    is_enabled.assert_called_once()
    console_print.assert_not_called()


@pytest.mark.skipif(
    not sys.platform.startswith("win"), reason="Requires Windows Registry"
)
@mock.patch("cumulusci.cli.utils.win32_long_paths_enabled")
@mock.patch("rich.console.Console.print")
def test_longpath_warn_on_win(console_print, is_enabled):
    is_enabled.return_value = False

    utils.warn_if_no_long_paths()

    is_enabled.assert_called_once()
    console_print.assert_called_once_with(utils.WIN_LONG_PATH_WARNING)



import functools
import re
import time
from pathlib import Path

import robot.api as robot_api
from robot.libraries.BuiltIn import BuiltIn
from selenium.common.exceptions import (
    ElementNotInteractableException,
    NoSuchWindowException,
    StaleElementReferenceException,
    WebDriverException,
)
from selenium.webdriver.remote.command import Command
from SeleniumLibrary.errors import ElementNotFound


def set_pdb_trace(pm=False):  # pragma: no cover
    """Start the Python debugger when robotframework is running.

    This makes sure that pdb can use stdin/stdout even though
    robotframework has redirected I/O.
    """
    import pdb
    import sys

    for attr in ("stdin", "stdout", "stderr"):
        setattr(sys, attr, getattr(sys, "__%s__" % attr))
    if pm:
        # Post-mortem debugging of an exception
        pdb.post_mortem()
    else:
        pdb.set_trace()


# This is a list of user actions that are likely to trigger
# Aura actions and/or XHRs. We'll add a step to wait for
# in-flight XHRs to complete after these commands.
COMMANDS_INVOKING_ACTIONS = {Command.CLICK_ELEMENT}


# This script waits for a) Aura to be available and b)
# any in-flight Aura XHRs to be complete.
# We only do this if the page uses Aura, as determined by looking for
# id="auraAppcacheProgress" in the DOM.
# It would be nice if we could inject the function when the page loads
# and then just call it after commands, but I was having trouble
# getting webdriver to add it to the window scope.
WAIT_FOR_AURA_SCRIPT = """
done = arguments[0];
if (document.getElementById('auraAppcacheProgress')) {
    var waitForXHRs = function() {
        if (window.$A && !window.$A.clientService.inFlightXHRs()) {
            done();
        } else {
            setTimeout(waitForXHRs, 100);
        }
    }
    setTimeout(waitForXHRs, 0);
} else {
    done();
}
"""


ALWAYS_RETRY_EXCEPTIONS = (
    ElementNotFound,
    ElementNotInteractableException,
    StaleElementReferenceException,
    NoSuchWindowException,
)


class RetryingSeleniumLibraryMixin(object):

    debug = False

    @property
    def selenium(self):
        selenium = self.builtin.get_library_instance("SeleniumLibrary")

        # Patch the selenium webdriver to add our own functionality
        # to improve stability.
        if not getattr(selenium.driver, "_cumulus_patched", False):
            orig_execute = selenium.driver.execute

            def execute(driver_command, params=None):
                try:
                    if self.retry_selenium:
                        # Retry certain failed commands once
                        result = self.selenium_execute_with_retry(
                            orig_execute, driver_command, params
                        )
                    else:
                        result = orig_execute(driver_command, params)
                except Exception:
                    if driver_command != Command.SCREENSHOT:
                        safe_screenshot()
                    if self.debug:
                        self.selenium.log_source()
                    raise

                # Run the "wait for aura" script after commands that are
                # likely to invoke async actions.
                if driver_command in COMMANDS_INVOKING_ACTIONS:
                    self.wait_for_aura()
                return result

            selenium.driver.execute = execute
            selenium.driver._cumulus_patched = True

        return selenium

    def selenium_execute_with_retry(self, execute, command, params):
        """Run a single selenium command and retry once.

        The retry happens for certain errors that are likely to be resolved
        by retrying.
        """
        try:
            return execute(command, params)
        except Exception as e:
            if isinstance(e, ALWAYS_RETRY_EXCEPTIONS) or (
                isinstance(e, WebDriverException)
                and "Other element would receive the click" in str(e)
            ):
                # Retry
                self.builtin.log("Retrying {} command".format(command), level="WARN")
                time.sleep(2)
                return execute(command, params)
            else:
                raise

    def wait_for_aura(self):
        """Run the WAIT_FOR_AURA_SCRIPT.

        This script polls Aura via $A in Javascript to determine when
        all in-flight XHTTP requests have completed before continuing.
        """
        try:
            self.selenium.driver.execute_async_script(WAIT_FOR_AURA_SCRIPT)
        except Exception:
            pass


def selenium_retry(target=None, retry=True):
    """Decorator to turn on automatic retries of flaky selenium failures.

    Decorate a robotframework library class to turn on retries for all
    selenium calls from that library::

        @selenium_retry
        class MyLibrary(object):

            # Decorate a method to turn it back off for that method
            @selenium_retry(False)
            def some_keyword(self):
                self.selenium.click_button('foo')

    Or turn it off by default but turn it on for some methods
    (the class-level decorator is still required)::

        @selenium_retry(False)
        class MyLibrary(object):

            @selenium_retry(True)
            def some_keyword(self):
                self.selenium.click_button('foo')
    """

    if isinstance(target, bool):
        # Decorator was called with a single boolean argument
        retry = target
        target = None

    def decorate(target):
        if isinstance(target, type):
            cls = target
            # Metaclass time.
            # We're going to generate a new subclass that:
            # a) mixes in RetryingSeleniumLibraryMixin
            # b) sets the initial value of `retry_selenium`
            return type(
                cls.__name__,
                (cls, RetryingSeleniumLibraryMixin),
                {"retry_selenium": retry, "__doc__": cls.__doc__},
            )
        func = target

        @functools.wraps(func)
        def run_with_retry(self, *args, **kwargs):
            # Set the retry setting and run the original function.
            old_retry = self.retry_selenium
            self.retry = retry
            try:
                return func(self, *args, **kwargs)
            finally:
                # Restore the previous value
                self.retry_selenium = old_retry

        run_with_retry.is_selenium_retry_decorator = True
        return run_with_retry

    if target is None:
        # Decorator is being used with arguments
        return decorate
    else:
        # Decorator was used without arguments
        return decorate(target)


def capture_screenshot_on_error(func):
    """Decorator for capturing a screenshot if a keyword throws an error

    This decorator will work for both SeleniumLibrary and Browser.

    If a screenshot cannot be taken for any reason, the screenshot error
    will be logged before the original exception is re-raised.

    For keywords using the Browser library, this will call the keyword
    "Take screenshot".

    For keywords using SeleniumLibrary, this decorator will call whatever
    keyword SeleniumLibrary is configured to call on failure. Normally
    that will be "Capture page screenshot".
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            safe_screenshot()
            raise

    return wrapper


def safe_screenshot():
    """Take a screenshot, but log and then ignore errors if the screenshot fails

    This function should work for both selenium and browser (playwright) libraries
    """

    libs = BuiltIn().get_library_instance(all=True)
    try:
        if "Browser" in libs:
            libs["Browser"].take_screenshot()
        elif "SeleniumLibrary" in libs:
            libs["SeleniumLibrary"].capture_page_screenshot()
    except Exception as e:
        BuiltIn().log(f"unable to capture screenshot: {e}")


def get_locator_module_name(version):
    """Return module name of locator file for the specified version

    If a file for the specified version can't be found, the locator
    file with the highest number will be returned. Passing in None
    effectively means "give me the latest version".
    """

    here = Path(__file__).parent
    if Path(here, f"locators_{version}.py").exists():
        # our work here is done.
        locator_name = f"locators_{version}"
    else:

        def by_num(filename):
            """Pull out the number from a filename, or return zero"""
            nums = re.findall(r"_(\d+)", str(filename))
            if nums:
                return int(nums[-1])
            return 0

        files = sorted(here.glob("locators_*.py"), key=by_num)
        locator_name = files[-1].stem
        robot_api.logger.warn(
            f"Locators for api {version} not found. Falling back to {locator_name} for salesforce locators"
        )

    locator_module_name = f"cumulusci.robotframework.{locator_name}"
    return locator_module_name



import contextlib
import fnmatch
import io
import math
import os
import re
import shutil
import sys
import tempfile
import textwrap
import zipfile
import functools
from datetime import datetime
from pathlib import Path
from typing import Union, Optional

import requests
import sarge

from cumulusci.vcs.models import AbstractRepo
from cumulusci.core.exceptions import CumulusCIException
from .xml import (  # noqa
    elementtree_parse_file,
    remove_xml_element,
    remove_xml_element_file,
    remove_xml_element_string,
)
from .ziputils import process_text_in_zipfile  # noqa
from .ziputils import zip_subfolder

CUMULUSCI_PATH = os.path.realpath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "../..")
)
META_XML_CLEAN_DIRS = ("classes/", "triggers/", "pages/", "aura/", "components/")
API_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
DATETIME_LEN = len("2018-08-07T16:00:56.000")

BREW_DEPRECATION_MSG = (
    "It looks like you have installed CumulusCI using brew."
    "This method of installation is no longer supported."
    "Please use the following to install CumulusCI with pipx:\n"
    "brew uninstall cumulusci-plus\nbrew install pipx\npipx ensurepath\npipx install cumulusci-plus"
)
PIP_UPDATE_CMD = "pip install --upgrade cumulusci-plus"
PIPX_UPDATE_CMD = "pipx upgrade cumulusci-plus"


def parse_api_datetime(value):
    """parse a datetime returned from the salesforce API.

    in python 3 we should just use a strptime %z, but until then we're just going
    to assert that its a fixed offset of +0000 since thats the observed behavior. getting
    python 2 to support fixed offset parsing is too complicated for what we need imo."""
    dt = datetime.strptime(value[0:DATETIME_LEN], API_DATE_FORMAT)
    offset_str = value[DATETIME_LEN:]
    assert offset_str in ["+0000", "Z"], "The Salesforce API returned a weird timezone."
    return dt


def find_replace(find, replace, directory, filePattern, logger=None, max=None):
    """Recursive find/replace.

    Walks through files matching `filePattern` within `directory`
    and does a string substitution of `find` with `replace`.
    """
    for path, dirs, files in os.walk(os.path.abspath(directory)):
        for filename in fnmatch.filter(files, filePattern):
            filepath = os.path.join(path, filename)
            with io.open(filepath, encoding="utf-8") as f:
                s = f.read()
            if max:
                s_updated = s.replace(find, replace, max)
            else:
                s_updated = s.replace(find, replace)
            if s != s_updated:
                if logger:
                    logger.info("Updating {}".format(filepath))
                with io.open(filepath, "w", encoding="utf-8") as f:
                    f.write(s_updated)


def find_replace_regex(find, replace, directory, filePattern, logger=None):
    """Recursive find/replace using a regular expression.

    Walks through files matching `filePattern` within `directory`
    and does a regex substitution of `find` with `replace`.
    """
    pattern = re.compile(find)
    for path, dirs, files in os.walk(os.path.abspath(directory)):
        for filename in fnmatch.filter(files, filePattern):
            filepath = os.path.join(path, filename)
            with io.open(filepath, encoding="utf-8") as f:
                s = f.read()
            s_updated = pattern.sub(replace, s)
            if s != s_updated:
                if logger:
                    logger.info("Updating {}".format(filepath))
                with io.open(filepath, "w", encoding="utf-8") as f:
                    f.write(s_updated)


def find_rename(find, replace, directory, logger=None):
    """Recursive find/replace within filenames.

    Walks through files within `directory`
    and renames files to replace `find` with `replace`.
    """
    for path, dirs, files in os.walk(os.path.abspath(directory)):
        for filename in files:
            filepath = os.path.join(path, filename)
            if logger:
                logger.info("Renaming {}".format(filepath))
            os.rename(filepath, os.path.join(path, filename.replace(find, replace)))


def remove_xml_element_directory(name, directory, file_pattern, logger=None):
    """Recursively walk a directory and remove XML elements"""
    for path, dirs, files in os.walk(os.path.abspath(directory)):
        for filename in fnmatch.filter(files, file_pattern):
            filepath = os.path.join(path, filename)
            remove_xml_element_file(name, filepath)


# backwards-compatibility aliases
findReplace = find_replace
findReplaceRegex = find_replace_regex
findRename = find_rename
removeXmlElement = remove_xml_element_directory


def download_extract_zip(url, target=None, subfolder=None, headers=None):
    if not headers:
        headers = {}
    resp = requests.get(url, headers=headers)
    zip_content = io.BytesIO(resp.content)
    zip_file = zipfile.ZipFile(zip_content)
    if subfolder:
        zip_file = zip_subfolder(zip_file, subfolder)
    if target:
        zip_file.extractall(target)
        return
    return zip_file


def download_extract_github(
    github_api, repo_owner, repo_name, subfolder=None, ref=None
):
    return download_extract_github_from_repo(
        github_api.repository(repo_owner, repo_name), subfolder, ref
    )


def download_extract_github_from_repo(github_repo, subfolder=None, ref=None):
    if not ref:
        ref = github_repo.default_branch
    zip_content = io.BytesIO()
    if not github_repo.archive("zipball", zip_content, ref=ref):
        raise CumulusCIException(
            f"Unable to download an archive of the Git ref {ref} from "
            f"{github_repo.full_name}. This can mean that the ref has "
            "not been pushed to the server, that CumulusCI's credential "
            "does not have permission to access it, or that your access "
            "is restricted by an IP address allow list."
        )
    zip_file = zipfile.ZipFile(zip_content)
    path = sorted(zip_file.namelist())[0]
    if subfolder:
        path = path + subfolder
    zip_file = zip_subfolder(zip_file, path)
    return zip_file


def download_extract_vcs_from_repo(
    vcs_repo: AbstractRepo, subfolder: Optional[str] = None, ref: Optional[str] = None
) -> zipfile.ZipFile:
    # Download a zip archive from a VCS repository.
    # If subfolder is specified, it will be extracted from the zip file.
    # If ref is specified, it will be used to download the archive.
    # If ref is not specified, the default branch of the repository will be used.
    # If the archive cannot be downloaded, a CumulusCIException will be raised.
    # The function returns a zipfile.ZipFile object containing the downloaded archive.
    if not ref:
        ref = vcs_repo.default_branch
    zip_content = io.BytesIO()
    if not vcs_repo.archive("zipball", zip_content, ref=ref):
        raise CumulusCIException(
            f"Unable to download an archive of the Git ref {ref} from "
            f"{vcs_repo.full_name}. This can mean that the ref has "
            "not been pushed to the server, that CumulusCI's credential "
            "does not have permission to access it, or that your access "
            "is restricted by an IP address allow list."
        )
    zip_file = zipfile.ZipFile(zip_content)
    path = sorted(zip_file.namelist())[0]

    root_folders_list = set(
        [fname.split("/")[0] for fname in sorted(zip_file.namelist())]
    )
    if len(root_folders_list) > 1:
        # If there are multiple root folders, we assume the zip contains all the repo content.
        path = ""

    if subfolder:
        path = path + subfolder
    zip_file = zip_subfolder(zip_file, path)
    return zip_file


def process_text_in_directory(path, process_file):
    """Process each file in a directory using the `process_file` function.

    `process_file` should be a function which accepts a filename and content as text
    and returns a (possibly modified) filename and content.  The file will be
    updated with the new content, and renamed if necessary.

    Files with content that cannot be decoded as UTF-8 will be skipped.
    """

    for path, dirs, files in os.walk(path):
        for orig_name in files:
            orig_path = os.path.join(path, orig_name)
            try:
                with open(orig_path, "r", encoding="utf-8") as f:
                    orig_content = f.read()
            except UnicodeDecodeError:
                # Probably a binary file; skip it
                continue
            new_name, new_content = process_file(orig_name, orig_content)
            new_path = os.path.join(path, new_name)
            if new_name != orig_name:
                os.rename(orig_path, new_path)
            with open(new_path, "w", encoding="utf-8") as f:
                f.write(new_content)


def inject_namespace(
    name,
    content,
    namespace=None,
    managed=None,
    filename_token=None,
    namespace_token=None,
    namespaced_org=None,
    logger=None,
):
    """Replaces %%%NAMESPACE%%% in file content and ___NAMESPACE___ in file name
    with either '' if no namespace is provided or 'namespace__' if provided.
    """

    # Handle namespace and filename tokens
    if not filename_token:
        filename_token = "___NAMESPACE___"
    if not namespace_token:
        namespace_token = "%%%NAMESPACE%%%"
    if managed is True and namespace:
        namespace_prefix = namespace + "__"
        namespace_dot_prefix = namespace + "."
    else:
        namespace_prefix = ""
        namespace_dot_prefix = ""

    namespace_dot_token = "%%%NAMESPACE_DOT%%%"

    # Handle tokens %%%NAMESPACED_ORG%%% and ___NAMESPACED_ORG___
    namespaced_org_token = "%%%NAMESPACED_ORG%%%"
    namespaced_org_file_token = "___NAMESPACED_ORG___"
    namespaced_org = namespace + "__" if namespaced_org else ""

    # Handle token %%%NAMESPACE_OR_C%%% for lightning components
    namespace_or_c_token = "%%%NAMESPACE_OR_C%%%"
    namespace_or_c = namespace if managed and namespace else "c"

    # Handle token %%%NAMESPACED_ORG_OR_C%%%
    namespaced_org_or_c_token = "%%%NAMESPACED_ORG_OR_C%%%"
    namespaced_org_or_c = namespace if namespaced_org else "c"

    orig_name = name
    prev_content = content
    content = content.replace(namespace_token, namespace_prefix)
    if logger and content != prev_content:
        logger.info(f'  {name}: Replaced {namespace_token} with "{namespace_prefix}"')

    prev_content = content
    content = content.replace(namespace_dot_token, namespace_dot_prefix)
    if logger and content != prev_content:
        logger.info(
            f'  {name}: Replaced {namespace_dot_token} with "{namespace_dot_prefix}"'
        )

    prev_content = content
    content = content.replace(namespace_or_c_token, namespace_or_c)
    if logger and content != prev_content:
        logger.info(
            f'  {name}: Replaced {namespace_or_c_token} with "{namespace_or_c}"'
        )

    if name == "package.xml":
        prev_content = content
        content = content.replace(filename_token, namespace_prefix)
        if logger and content != prev_content:
            logger.info(
                f'  {name}: Replaced {filename_token} with "{namespace_prefix}"'
            )

        # Also replace ___NAMESPACED_ORG___ tokens in package.xml
        prev_content = content
        content = content.replace(namespaced_org_file_token, namespaced_org)
        if logger and content != prev_content:
            logger.info(
                f'  {name}: Replaced {namespaced_org_file_token} with "{namespaced_org}"'
            )

    prev_content = content
    content = content.replace(namespaced_org_token, namespaced_org)
    if logger and content != prev_content:
        logger.info(
            f'  {name}: Replaced {namespaced_org_token} with "{namespaced_org}"'
        )

    prev_content = content
    content = content.replace(namespaced_org_or_c_token, namespaced_org_or_c)
    if logger and content != prev_content:
        logger.info(
            f'  {name}: Replaced {namespaced_org_or_c_token} with "{namespaced_org_or_c}"'
        )

    # Replace namespace token in file name
    name = name.replace(filename_token, namespace_prefix)
    name = name.replace(namespaced_org_file_token, namespaced_org)
    if logger and name != orig_name:
        logger.info(f"  {orig_name}: renamed to {name}")

    return name, content


def strip_namespace(name, content, namespace, logger=None):
    """Given a namespace, strips 'namespace__' from file name and content"""
    namespace_prefix = "{}__".format(namespace)
    lightning_namespace = "{}:".format(namespace)

    orig_content = content
    new_content = orig_content.replace(namespace_prefix, "")
    new_content = new_content.replace(lightning_namespace, "c:")
    name = name.replace(namespace_prefix, "")
    if orig_content != new_content and logger:
        logger.info(
            "  {file_name}: removed {namespace}".format(
                file_name=name, namespace=namespace_prefix
            )
        )
    return name, new_content


def tokenize_namespace(name, content, namespace, logger=None):
    """Given a namespace, replaces 'namespace__' with %%%NAMESPACE%%%
    in file content and ___NAMESPACE___ in file name
    """
    if not namespace:
        return name, content

    namespace_prefix = "{}__".format(namespace)
    lightning_namespace = "{}:".format(namespace)

    content = content.replace(namespace_prefix, "%%%NAMESPACE%%%")
    content = content.replace(lightning_namespace, "%%%NAMESPACE_OR_C%%%")
    name = name.replace(namespace_prefix, "___NAMESPACE___")

    return name, content


def zip_clean_metaxml(zip_src, logger=None):
    """Given a zipfile, cleans all ``*-meta.xml`` files in the zip for
    deployment by stripping all ``<packageVersions/>`` elements
    """
    zip_dest = zipfile.ZipFile(io.BytesIO(), "w", zipfile.ZIP_DEFLATED)
    changed = []
    for name in zip_src.namelist():
        content = zip_src.read(name)
        if name.startswith(META_XML_CLEAN_DIRS) and name.endswith("-meta.xml"):
            try:
                content.decode("utf-8")
            except UnicodeDecodeError:
                # if we cannot decode the content, it may be binary;
                # don't try and replace it.
                pass
            else:
                clean_content = remove_xml_element_string("packageVersions", content)
                if clean_content != content:
                    changed.append(name)
                    content = clean_content
        zip_dest.writestr(name, content)
    if changed and logger:
        logger.info(
            "Cleaned package versions from {} meta.xml files".format(len(changed))
        )
    zip_src.close()
    return zip_dest


def doc_task(task_name, task_config, project_config=None, org_config=None):
    """Document a (project specific) task configuration in RST format."""
    from cumulusci.core.utils import import_global

    doc = []
    doc.append(f".. _{task_name.replace('_', '-')}:\n")
    doc.append(f"{task_name}\n==========================================\n")
    doc.append(f"**Description:** {task_config.description}\n")
    doc.append(f"**Class:** {task_config.class_path}\n")

    task_class = import_global(task_config.class_path)

    if "task_docs" in task_class.__dict__:
        task_docs = textwrap.dedent(task_class.task_docs.strip("\n"))
        doc.append(task_docs)

    task_option_info = get_task_option_info(task_config, task_class)
    doc.append("Command Syntax\n------------------------------------------\n")
    command_syntax = get_command_syntax(task_name)
    doc.append(command_syntax)

    task_option_doc = create_task_options_doc(task_option_info)
    if task_option_doc:
        doc.append("Options\n------------------------------------------\n")
        doc.extend(task_option_doc)

    return "\n".join(doc)


def get_command_syntax(task_name):
    """Return an example command syntax string in .rst format"""
    return f"``$ cci task run {task_name}``\n\n"


def get_task_option_info(task_config, task_class):
    """Gets the the following info for each option in the task
    usage: example usage statement (i.e. -o name VALUE)
    required: True/False
    default: If a default value is present
    description: Description string provided on the task option
    option_type: A type string provided on the task option

    Returns list of option dicts with required at the front of the map
    """
    required_options = []
    optional_options = []
    defaults = task_config.options or {}

    for name, option in list(task_class.task_options.items()):
        usage = get_option_usage_string(name, option)
        required = True if option.get("required") else False
        default = defaults.get(name)
        description = option.get("description")
        option_type = option.get("type")

        info = {
            "usage": usage,
            "name": name,
            "required": required,
            "default": default,
            "description": description,
            "option_type": option_type,
        }
        if required:
            required_options.append(info)
        else:
            optional_options.append(info)

    return [*required_options, *optional_options]


def get_option_usage_string(name, option):
    """Returns a usage string if one exists
    else creates a usage string in the form of:

        --option-name OPTIONNAME
    """
    usage_str = option.get("usage")
    if not usage_str:
        usage_str = f"--{name} {name.replace('_', '').upper()}"
    return usage_str


def create_task_options_doc(task_options):
    """Generate the 'Options' section for a given tasks documentation"""
    doc = []
    for option in task_options:
        usage_str = option.get("usage")
        if usage_str:
            doc.append(f"\n``{usage_str}``")

        description = option.get("description")
        if description:
            doc.append(f"\n\t {description}")

        default = option.get("default")
        if default:
            doc.append(f"\n\t Default: {default}")
        elif option.get("required"):
            doc.append("\n *Required*")
        else:
            doc.append("\n *Optional*")

        option_type = option.get("option_type")
        if option_type:
            doc.append(f"\n\t Type: {option_type}")

    return doc


def flow_ref_title_and_intro(intro_blurb):
    return f"""Flow Reference
==========================================
\n{intro_blurb}

"""


def document_flow(flow_name, description, flow_coordinator, additional_info=None):
    """Document (project specific) flow configurations in RST format"""
    doc = []

    doc.append(f".. _{flow_name.replace('_', '-')}:\n")
    doc.append(f"{flow_name}\n{'^' * len(flow_name)}\n")
    doc.append(f"**Description:** {description}\n")

    if additional_info:
        doc.append(additional_info)

    doc.append("**Flow Steps**\n")
    doc.append(".. code-block:: console\n")
    flow_step_lines = flow_coordinator.get_flow_steps(for_docs=True)
    # extra indent beneath code-block and finish with pipe for extra space afterwards
    flow_step_lines = [f"\t{line}" for line in flow_step_lines]
    # fix when clauses
    lines = []
    for line in flow_step_lines:
        if line.startswith("when"):
            line = f"\t\t{line}"
        lines.append(line)
    doc.extend(lines)

    return "\n".join(doc)


def package_xml_from_dict(items, api_version, package_name=None):
    lines = []

    # Print header
    lines.append('<?xml version="1.0" encoding="UTF-8"?>')
    lines.append('<Package xmlns="http://soap.sforce.com/2006/04/metadata">')

    # Include package name if specified
    if package_name:
        lines.append("    <fullName>{}</fullName>".format(package_name))

    # Print types sections
    for md_type, members in sorted(items.items()):
        members.sort()
        lines.append("    <types>")
        for member in members:
            lines.append("        <members>{}</members>".format(member))
        lines.append("        <name>{}</name>".format(md_type))
        lines.append("    </types>")

    # Print footer
    lines.append("    <version>{0}</version>".format(api_version))
    lines.append("</Package>")

    return "\n".join(lines)


@contextlib.contextmanager
def cd(path):
    """Context manager that changes to another directory"""
    if not path:
        yield
        return
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)


@contextlib.contextmanager
def temporary_dir(chdir=True):
    """Context manager that creates a temporary directory and chdirs to it.

    When the context manager exits it returns to the previous cwd
    and deletes the temporary directory.
    """
    d = tempfile.mkdtemp()
    try:
        with contextlib.ExitStack() as stack:
            if chdir:
                stack.enter_context(cd(d))
            yield d
    finally:
        if os.path.exists(d):
            try:
                shutil.rmtree(d)
            except Exception as e:  # pragma: no cover
                import logging  # needs to be local or cumulusci.utils.logging gets picked up

                logging.getLogger(__file__).warn(
                    f"Cannot remove temporary directory {d} because: {e}"
                )


def touch(path):
    """Ensure a file exists."""
    with open(path, "a"):
        pass


def in_directory(filepath, dirpath):
    """Returns a boolean for whether filepath is contained in dirpath.

    Normalizes the paths (e.g. resolving symlinks and ..)
    so this is the safe way to make sure a user-configured path
    is located inside the user's project repo.
    """
    filepath = os.path.realpath(filepath)
    dirpath = os.path.realpath(dirpath)
    return filepath == dirpath or filepath.startswith(os.path.join(dirpath, ""))


def log_progress(
    iterable,
    logger,
    batch_size=10000,
    progress_message="Processing... ({})",
    done_message="Done! (Total: {})",
):
    """Log progress while iterating."""
    i = 0
    for x in iterable:
        yield x
        i += 1
        if not i % batch_size:
            logger.info(progress_message.format(i))
    logger.info(done_message.format(i))


def random_alphanumeric_underscore(length):
    import secrets

    # Ensure the string is the right length
    byte_length = math.ceil((length * 3) / 4)
    return secrets.token_urlsafe(byte_length).replace("-", "_")[:length]


def get_cci_upgrade_command():
    deprecated_install_paths = ["cellar", "linuxbrew"]
    for path in deprecated_install_paths:
        if path in CUMULUSCI_PATH.lower():
            return BREW_DEPRECATION_MSG

    return PIPX_UPDATE_CMD if "pipx" in CUMULUSCI_PATH.lower() else PIP_UPDATE_CMD


def convert_to_snake_case(content):
    s1 = re.sub("([^_])([A-Z][a-z]+)", r"\1_\2", content)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def get_git_config(config_key):
    p = sarge.Command(
        sarge.shell_format('git config --get "{0!s}"', config_key),
        stderr=sarge.Capture(buffer_size=-1),
        stdout=sarge.Capture(buffer_size=-1),
        shell=True,
    )
    p.run()
    config_value = (
        io.TextIOWrapper(p.stdout, encoding=sys.stdout.encoding).read().strip()
    )

    return config_value if config_value and not p.returncode else None


def update_tree(src: Union[str, Path], dest: Union[str, Path]):
    """
    Copies files from src to dest, same as distutils.copy_tree(update=1).

    Copies the entire directory tree from src to dest. If dest exists, only
    copies files that are newer in src than in dest, or files that don't exist
    in dest.

    Args:
        src (Union[str, Path]): The source directory to copy files from.
        dest (Union[str, Path]): The destination directory to copy files to.
    """
    src_path = Path(src)
    dest_path = Path(dest)
    if not dest_path.exists():
        shutil.copytree(src_path, dest_path)
    else:
        for src_dir in src_path.rglob("*"):
            if src_dir.is_file():
                dest_file = dest_path / src_dir.relative_to(src_path)
                if (
                    not dest_file.exists()
                    or src_dir.stat().st_mtime - dest_file.stat().st_mtime > 1
                ):
                    dest_file.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy2(src_dir, dest_file)


def filter_namelist(includes, namelist):
    """
    Filter a zipfile namelist, handling any included directory filenames missing
    a trailing slash.
    """
    included_dirs = []
    zip_dirs = [filename.rstrip("/") for filename in namelist if filename.endswith("/")]

    for name in includes:
        if name.endswith("/"):
            included_dirs.append(name)
        elif name in zip_dirs:
            # append a trailing slash to avoid partial matches
            included_dirs.append(name + "/")

    return list(
        {
            name
            for name in namelist
            if name.startswith(tuple(included_dirs)) or name in includes
        }
    )


@functools.lru_cache(50)
def get_tasks_with_options(project_config, frozen_options: frozenset):
    coordinator_opts = {}
    options = dict(frozen_options)
    for task_config in project_config.list_tasks():
        if any(
            key in options.keys()
            for key in project_config.get_task(task_config["name"])
            .get_class()
            .task_options.keys()
        ):
            coordinator_opts[task_config["name"]] = options
    return coordinator_opts



import io
import os
from contextlib import contextmanager, nullcontext
from http.client import HTTPMessage
from logging import getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import pytest
import yaml
from pytest import fixture
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from cumulusci.core.dependencies.utils import TaskContext
from cumulusci.core.github import get_github_api
from cumulusci.salesforce_api.org_schema_models import Base
from cumulusci.tasks.bulkdata.tests.integration_test_utils import (
    ensure_accounts,
    ensure_records,
)
from cumulusci.tasks.salesforce.tests.util import create_task_fixture
from cumulusci.tests.pytest_plugins.pytest_sf_vcr import salesforce_vcr, vcr_config
from cumulusci.tests.util import DummyKeychain, DummyOrgConfig, mock_env

ensure_accounts = ensure_accounts
ensure_records = ensure_records


@fixture(scope="session", autouse=True)
def mock_sleep():
    """Patch time.sleep to avoid delays in unit tests"""
    with mock.patch("time.sleep"):
        yield


class MockHttpResponse(mock.Mock):
    def __init__(self, status):
        super(MockHttpResponse, self).__init__()
        self.status = status
        self.strict = 0
        self.version = 0
        self.reason = None
        self.msg = HTTPMessage(io.BytesIO())
        self.closed = True

    def read(self):  # pragma: no cover
        return b""

    def isclosed(self):
        return self.closed


@fixture
def gh_api():
    return get_github_api("TestOwner", "TestRepo")


@fixture(scope="class", autouse=True)
def restore_cwd():
    d = os.getcwd()
    try:
        yield
    finally:
        os.chdir(d)


@fixture
def mock_http_response():
    def _make_response(status):
        return MockHttpResponse(status)

    return _make_response


@fixture(scope="function")
def fallback_org_config():
    def fallback_org_config():
        return DummyOrgConfig(
            name="pytest_sf_orgconnect_dummy_org_config", keychain=DummyKeychain()
        )

    return fallback_org_config


vcr_config = fixture(vcr_config, scope="module")
vcr = fixture(salesforce_vcr, scope="module")

create_task_fixture = fixture(create_task_fixture, scope="function")


# TODO: This should also chdir to a temp directory which
#       can represent the repo-root, but that will require
#       test case changes.
@pytest.fixture(autouse=True)
def patch_home_and_env(request):
    "Patch the default home directory and $HOME environment for all tests at once."

    use_real_env = request.node.get_closest_marker("use_real_env")

    with TemporaryDirectory(prefix="fake_home_") as home:
        if use_real_env:
            mock_env_cm = nullcontext()
        else:
            mock_env_cm = mock_env(home)

        with mock_env_cm:
            Path(home, ".cumulusci").mkdir()
            Path(home, ".cumulusci/cumulusci.yml").touch()
            yield


@pytest.fixture()
def temp_db():
    with TemporaryDirectory() as t:

        @contextmanager
        def open_db():
            engine = create_engine(f"sqlite:///{t}/tempfile.db")
            with engine.connect() as connection:
                Session = sessionmaker(bind=connection)
                Base.metadata.bind = engine
                Base.metadata.create_all()
                session = Session()
                yield connection, Base.metadata, session

        yield open_db


@pytest.fixture()
def delete_data_from_org(create_task):
    def delete_data_from_org(object_names):
        from cumulusci.tasks.bulkdata.delete import DeleteData

        t = create_task(DeleteData, {"objects": object_names})
        assert t.org_config.scratch
        t()

    return delete_data_from_org


@pytest.fixture(scope="session")
def cumulusci_test_repo_root():
    return Path(__file__).parent.parent


@pytest.fixture(scope="session")
def global_describe(cumulusci_test_repo_root):
    global_describe_file = (
        cumulusci_test_repo_root
        / "cumulusci/tests/shared_cassettes/GET_sobjects_Global_describe.yaml"
    )
    with global_describe_file.open() as f:
        data = yaml.safe_load(yaml.safe_load(f)["response"]["body"]["string"])

    def global_describe_specific_sobjects(sobjects: int = None):
        if sobjects is None:  # pragma: no cover
            subset = data.copy()
        elif isinstance(sobjects, int):
            subset = data.copy()
            subset["sobjects"] = subset["sobjects"][0:sobjects]
        elif isinstance(sobjects, (list, tuple)):  # pragma: no cover
            raise NotImplementedError(
                "We could implement a by-name subsetting here when we need it."
            )
        return subset

    return global_describe_specific_sobjects


@pytest.fixture(scope="session")
def shared_vcr_cassettes(cumulusci_test_repo_root):
    return Path(cumulusci_test_repo_root / "cumulusci/tests/shared_cassettes")


@pytest.fixture
def task_context(org_config, project_config):
    return TaskContext(
        org_config=org_config, project_config=project_config, logger=getLogger()
    )



from cumulusci.tasks.github.merge import MergeBranch
from cumulusci.tasks.github.pull_request import PullRequests
from cumulusci.tasks.github.release import CreateRelease
from cumulusci.tasks.github.release_report import ReleaseReport
from cumulusci.tasks.github.tag import CloneTag
from cumulusci.tasks.github.commit_status import GetPackageDataFromCommitStatus
from cumulusci.tasks.github.publish import PublishSubtree

__all__ = (
    "MergeBranch",
    "PullRequests",
    "CreateRelease",
    "ReleaseReport",
    "CloneTag",
    "GetPackageDataFromCommitStatus",
    "PublishSubtree",
)



import json

import sarge

from cumulusci.core.exceptions import SalesforceDXException
from cumulusci.core.sfdx import sfdx
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.utils.options import CCIOptions, Field


class SfDataCommands(BaseSalesforceApiTask):
    class Options(CCIOptions):
        json_output: bool = Field(
            None, description="Whether to return the result as a JSON object"
        )
        api_version: str = Field(None, description="API version to use for the command")
        flags_dir: str = Field(None, description="Import flag values from a directory")

    parsed_options: Options

    def _init_task(self):
        super()._init_task()

    def _init_options(self, kwargs):
        self.args = []
        self.data_command = "data "
        super()._init_options(kwargs)
        if self.parsed_options.flags_dir:
            self.args.extend(["--flags-dir ", self.parsed_options.flags_dir])
        if self.parsed_options.json_output:
            self.args.extend(["--json"])
        if self.parsed_options.api_version:
            self.args.extend(["--api_version", self.parsed_options.api_version])

    def _run_task(self):
        self.return_values = {}

        self.p: sarge.Command = sfdx(
            self.data_command,
            log_note="Running data command",
            args=self.args,
            check_return=True,
            username=self.org_config.username,
        )

        if self.parsed_options.json_output:
            self.return_values = self._load_json_output(self.p)

        for line in self.p.stdout_text:
            self.logger.info(line)

        for line in self.p.stderr_text:
            self.logger.error(line)

    def _load_json_output(self, p: sarge.Command, stdout: str = None):
        try:
            stdout = stdout or p.stdout_text.read()
            return json.loads(stdout)
        except json.decoder.JSONDecodeError:
            raise SalesforceDXException(
                f"Failed to parse the output of the {self.data_command} command"
            )


class SfDataToolingAPISupportedCommands(SfDataCommands):
    class Options(SfDataCommands.Options):
        use_tooling_api: bool = Field(
            None,
            description="Use Tooling API so you can run queries on Tooling API objects.",
        )


class DataQueryTask(SfDataToolingAPISupportedCommands):
    class Options(SfDataToolingAPISupportedCommands.Options):
        query: str = Field(None, description="SOQL query to execute")
        file: str = Field(None, description="File that contains the SOQL query")
        all_rows: bool = Field(
            None,
            description="Include deleted records. By default, deleted records are not returned.",
        )
        result_format: str = Field(
            None,
            description="Format to display the results; the --json_output flag overrides this flag. Permissible values are: human, csv, json.",
        )
        output_file: str = Field(
            None,
            description="File where records are written; only CSV and JSON output formats are supported.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "query"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.parsed_options.query:
            self.args.extend(["--query", self.parsed_options.query])
        if self.parsed_options.file:
            self.args.extend(["--file", self.parsed_options.file])
        if self.parsed_options.all_rows:
            self.args.extend(["--all-rows", self.parsed_options.all_rows])
        if self.parsed_options.result_format:
            self.args.extend(["--result-format", self.parsed_options.result_format])
        if self.parsed_options.output_file:
            self.args.extend(["--output-file", self.parsed_options.output_file])

    def _run_task(self):
        super()._run_task()

        if self.parsed_options.json_output:
            self.logger.info(self.return_values)


class DataCreateRecordTask(SfDataToolingAPISupportedCommands):
    class Options(SfDataToolingAPISupportedCommands.Options):
        sobject: str = Field(
            ...,
            description="API name of the Salesforce or Tooling API object that you're inserting a record into.",
        )
        values: str = Field(
            ...,
            description="Values for the flags in the form <fieldName>=<value>, separate multiple pairs with spaces.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "create record"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.options.get("sobject"):
            self.args.extend(["--sobject", self.options.get("sobject")])
        if self.options.get("values"):
            self.args.extend(["--values", self.options.get("values")])

    def _run_task(self):
        return super()._run_task()


class DataDeleteRecordTask(SfDataToolingAPISupportedCommands):
    class Options(SfDataToolingAPISupportedCommands.Options):
        sobject: str = Field(
            ...,
            description="API name of the Salesforce or Tooling API object that you're deleting a record from.",
        )
        record_id: str = Field(None, description="ID of the record you’re deleting.")
        where: str = Field(
            None,
            description="List of <fieldName>=<value> pairs that identify the record you want to delete.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "delete record"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.options.get("sobject"):
            self.args.extend(["--sobject", self.options.get("sobject")])
        if self.options.get("record_id"):
            self.args.extend(["--record-id", self.options.get("record_id")])
        if self.options.get("where"):
            self.args.extend(["--where", self.options.get("where")])

    def _run_task(self):
        return super()._run_task()


class DataUpdateRecordTask(SfDataToolingAPISupportedCommands):
    class Options(SfDataToolingAPISupportedCommands.Options):
        sobject: str = Field(
            ...,
            description="API name of the Salesforce or Tooling API object that you're updating a record from.",
        )
        record_id: str = Field(None, description="ID of the record you’re updating.")
        where: str = Field(
            None,
            description="List of <fieldName>=<value> pairs that identify the record you want to update.",
        )
        values: str = Field(
            ...,
            description="Values for the flags in the form <fieldName>=<value>, separate multiple pairs with spaces.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "update record"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.parsed_options.sobject:
            self.args.extend(["--sobject", self.parsed_options.sobject])
        if self.parsed_options.record_id:
            self.args.extend(["--record-id", self.parsed_options.record_id])
        if self.parsed_options.where:
            self.args.extend(["--where", self.parsed_options.where])
        if self.parsed_options.values:
            self.args.extend(["--values", self.parsed_options.values])

    def _run_task(self):
        return super()._run_task()


class DataGetRecordTask(SfDataToolingAPISupportedCommands):
    class Options(SfDataToolingAPISupportedCommands.Options):
        sobject: str = Field(
            ...,
            description="API name of the Salesforce or Tooling API object that you're fetching a record from.",
        )
        record_id: str = Field(None, description="ID of the record you’re fetching.")
        where: str = Field(
            None,
            description="List of <fieldName>=<value> pairs that identify the record you want to fetch.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "get record"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.parsed_options.sobject:
            self.args.extend(["--sobject", self.parsed_options.sobject])
        if self.parsed_options.record_id:
            self.args.extend(["--record-id", self.parsed_options.record_id])
        if self.parsed_options.where:
            self.args.extend(["--where", self.parsed_options.where])

    def _run_task(self):
        return super()._run_task()


class DataQueryResumeTask(SfDataToolingAPISupportedCommands):
    class Options(SfDataToolingAPISupportedCommands.Options):
        bulk_query_id: str = Field(
            ...,
            description="The 18-character ID of the bulk query to resume.",
        )
        result_format: str = Field(
            None,
            description="Format to display the results; the --json_output flag overrides this flag. Permissible values are: human, csv, json.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "query resume"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.parsed_options.bulk_query_id:
            self.args.extend(["--bulk-query-id", self.parsed_options.bulk_query_id])
        if self.parsed_options.result_format:
            self.args.extend(["--result-format", self.parsed_options.result_format])

    def _run_task(self):
        return super()._run_task()


class DataDeleteBulkTask(SfDataCommands):
    class Options(SfDataCommands.Options):
        sobject: str = Field(
            ...,
            description="The API name of the object for the bulk job.",
        )
        file: str = Field(
            ...,
            description="The path to the CSV file that contains the IDs of the records to delete.",
        )
        wait: int = Field(
            None,
            description="The number of minutes to wait for the command to complete.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "delete bulk"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.parsed_options.sobject:
            self.args.extend(["--sobject", self.parsed_options.sobject])
        if self.parsed_options.file:
            self.args.extend(["--file", self.parsed_options.file])
        if self.parsed_options.wait:
            self.args.extend(["--wait", str(self.parsed_options.wait)])

    def _run_task(self):
        return super()._run_task()


class DataUpsertBulkTask(SfDataCommands):
    class Options(SfDataCommands.Options):
        sobject: str = Field(
            ...,
            description="The API name of the object for the bulk job.",
        )
        file: str = Field(
            ...,
            description="The path to the CSV file that contains the records to upsert.",
        )
        external_id_field: str = Field(
            ...,
            description="The API name of the external ID field for the upsert.",
        )
        wait: int = Field(
            None,
            description="The number of minutes to wait for the command to complete.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "upsert bulk"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.parsed_options.sobject:
            self.args.extend(["--sobject", self.parsed_options.sobject])
        if self.parsed_options.file:
            self.args.extend(["--file", self.parsed_options.file])
        if self.parsed_options.external_id_field:
            self.args.extend(
                ["--external-id-field", self.parsed_options.external_id_field]
            )
        if self.parsed_options.wait:
            self.args.extend(["--wait", str(self.parsed_options.wait)])

    def _run_task(self):
        return super()._run_task()


class DataImportTreeTask(SfDataCommands):
    class Options(SfDataCommands.Options):
        files: list = Field(
            None,
            description="A list of paths to sObject Tree API plan definition files.",
        )
        plan: str = Field(
            None,
            description="The path to a plan definition file.",
        )
        content_type_map: str = Field(
            None,
            description="A mapping of file extensions to content types.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "import tree"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.parsed_options.files:
            self.args.extend(["--files", ",".join(self.parsed_options.files)])
        if self.parsed_options.plan:
            self.args.extend(["--plan", self.parsed_options.plan])
        if self.parsed_options.content_type_map:
            self.args.extend(
                ["--content-type-map", self.parsed_options.content_type_map]
            )

    def _run_task(self):
        return super()._run_task()


class DataExportTreeTask(SfDataCommands):
    class Options(SfDataCommands.Options):
        query: str = Field(
            ...,
            description="A SOQL query that retrieves the records you want to export.",
        )
        plan: bool = Field(
            False,
            description="Generate a plan definition file.",
        )
        prefix: str = Field(
            None,
            description="The prefix for the exported data files.",
        )
        output_dir: str = Field(
            None,
            description="The directory to store the exported files.",
        )

    def _init_task(self):
        super()._init_task()
        self.data_command += "export tree"

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        if self.parsed_options.query:
            self.args.extend(["--query", self.parsed_options.query])
        if self.parsed_options.plan:
            self.args.extend(["--plan"])
        if self.parsed_options.prefix:
            self.args.extend(["--prefix", self.parsed_options.prefix])
        if self.parsed_options.output_dir:
            self.args.extend(["--output-dir", self.parsed_options.output_dir])

    def _run_task(self):
        return super()._run_task()



import base64
import io
import os
import zipfile
from unittest import mock

import pytest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.flowrunner import StepSpec
from cumulusci.core.source_transforms.transforms import CleanMetaXMLTransform
from cumulusci.tasks.salesforce import Deploy
from cumulusci.utils import temporary_dir, touch

from .util import create_task


class TestDeploy:
    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_get_api(self, mock_org_config, rest_deploy):
        with temporary_dir() as path:
            touch("package.xml")
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "namespace_tokenize": "ns",
                    "namespace_inject": "ns",
                    "namespace_strip": "ns",
                    "unmanaged": True,
                    "rest_deply": rest_deploy,
                },
            )

            api = task._get_api()
            zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(api.package_zip)), "r")
            assert "package.xml" in zf.namelist()
            zf.close()

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    def test_create_api_object(self, mock_org_config):
        with temporary_dir() as path:
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "namespace_tokenize": "ns",
                    "namespace_inject": "ns",
                    "namespace_strip": "ns",
                    "unmanaged": True,
                    "collision_check": True,
                },
            )
            package_xml = """<?xml version="1.0" encoding="UTF-8"?>
    <Package xmlns="http://soap.sforce.com/2006/04/metadata">
        <types>
            <members>Delivery__c.Supplier__c</members>
            <members>Delivery__c.Scheduled_Date__c</members>
            <name>CustomField</name>
        </types>
        <types>
            <name>CustomObject</name>
        </types>
        <types>
            <name>CustomTab</name>
        </types>
        <types>
            <name>Layout</name>
        </types>
        <types>
            <name>ListView</name>
        </types>
        <version>58.0</version>
    </Package>"""
            expected_package_xml = """<types><members>Delivery__c.Supplier__c</members><members>Delivery__c.Scheduled_Date__c</members><name>CustomField</name></types><types><name>CustomObject</name></types><types><name>CustomTab</name></types><types><name>Layout</name></types><types><name>ListView</name></types><version>58.0</version>    """
            api_object = task._create_api_object(package_xml, "58.0")

            assert api_object.package_xml.split() == expected_package_xml.split()
            assert api_object.api_version == "58.0"

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    def test_collision_check_positive(self, mock_org_config):
        with temporary_dir() as path:
            touch("package.xml")
            with open("package.xml", "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
    <Package xmlns="http://soap.sforce.com/2006/04/metadata">
        <types>
            <members>Delivery__c.Supplier__c</members>
            <members>Delivery__c.Scheduled_Date__c</members>
            <name>CustomField</name>
        </types>
        <types>
            <name>CustomObject</name>
        </types>
        <types>
            <name>CustomTab</name>
        </types>
        <types>
            <name>Layout</name>
        </types>
        <types>
            <name>ListView</name>
        </types>
        <version>58.0</version>
    </Package>"""
                )
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "namespace_tokenize": "ns",
                    "namespace_inject": "ns",
                    "namespace_strip": "ns",
                    "unmanaged": True,
                    "collision_check": True,
                },
            )

            task._create_api_object = mock.Mock(
                return_value=mock.Mock(
                    _get_response=mock.Mock(
                        return_value=mock.Mock(
                            content='<?xml version="1.0" encoding="UTF-8"?> <soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns="http://soap.sforce.com/2006/04/metadata"><soapenv:Body><checkRetrieveStatusResponse><result><done>true</done><fileProperties><createdById>0051m0000069dpxAAA</createdById><createdByName>User User</createdByName><createdDate>2023-10-09T08:47:44.875Z</createdDate><fileName>unpackaged/package.xml</fileName><fullName>unpackaged/package.xml</fullName><id></id><lastModifiedById>0051m0000069dpxAAA</lastModifiedById><lastModifiedByName>User User</lastModifiedByName><lastModifiedDate>2023-10-09T08:47:44.875Z</lastModifiedDate><manageableState>unmanaged</manageableState><type>Package</type></fileProperties><id>09S1m000001EKBkEAO</id><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery_Item__c.Food_Expiration_Date__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery__c.Status__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery_Item__c.Food_Storage__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery_Item__c.Delivery__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery__c.Scheduled_Date__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery__c.Supplier__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomObject&apos; named &apos;Delivery__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomObject&apos; named &apos;Delivery_Item__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomTab&apos; named &apos;Delivery__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;Layout&apos; named &apos;Delivery_Item__c-Delivery Item Layout&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;Layout&apos; named &apos;Delivery__c-Delivery Layout&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;ListView&apos; named &apos;Delivery__c.All&apos; cannot be found</problem></messages><status>Succeeded</status><success>true</success><zipFile>UEsDBBQACAgIAPZFSVcAAAAAAAAAAAAAAAAWAAAAdW5wYWNrYWdlZC9wYWNrYWdlLnhtbKVTTU/DMAy991dUva8pCNCE0kyIUQkJCaQNrlGami2QNFXjwvrvyT7KuFTraE7Ji5/fsy3T2cbo8Atqp2yZRhdxEoVQSluocpVGr8tsMo1mLKAvQn6KFYQ+unRptEasbglxVlSxe7e1hFhaQy6T5IYkV8QAikKgiFgQ+kOxrcDt77u3AZN7STYHrbx2yx8RDOcy/gW4pKSLGsDLrC34w6ZStUBfCJ8LhH/lWKCtfZ1DuZ62kGsoGg3FWaJbIgps3DmEpqq0grqHUgoD7L5xaE2mQBeU7JD9BMifEQwex0BjJ8w85x8gcYSbkwpLkY8vdtIB4RYIn0RrGxzYgSO3n7bz132P6EV8p3V/fuXwTcF3v8Jh1dn1NE4o6V4BJYcNZ8EPUEsHCGtr2HwiAQAAEwQAAFBLAQIUABQACAgIAPZFSVdra9h8IgEAABMEAAAWAAAAAAAAAAAAAAAAAAAAAAB1bnBhY2thZ2VkL3BhY2thZ2UueG1sUEsFBgAAAAABAAEARAAAAGYBAAAAAA==</zipFile></result></checkRetrieveStatusResponse></soapenv:Body></soapenv:Envelope>'
                        )
                    )
                )
            )

            is_collision, xml_map = task._collision_check(path)

            assert is_collision is False
            print(xml_map)
            assert xml_map == {
                "CustomObject": [],
                "CustomTab": [],
                "Layout": [],
                "ListView": [],
            }

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    def test_collision_check_negative(self, mock_org_config):

        with temporary_dir() as path:
            touch("package.xml")
            with open("package.xml", "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
    <Package xmlns="http://soap.sforce.com/2006/04/metadata">
       <types>
            <members>Delivery__c.Supplier__c</members>
            <members>Delivery__c.Scheduled_Date__c</members>
            <name>CustomField</name>
        </types>
        <types>
            <name>CustomObject</name>
        </types>
        <types>
            <name>CustomTab</name>
        </types>
        <types>
            <name>Layout</name>
        </types>
        <types>
            <name>ListView</name>
        </types>
        <version>58.0</version>
    </Package>"""
                )
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "namespace_tokenize": "ns",
                    "namespace_inject": "ns",
                    "namespace_strip": "ns",
                    "unmanaged": True,
                    "collision_check": True,
                },
            )

            task._create_api_object = mock.Mock(
                return_value=mock.Mock(
                    _get_response=mock.Mock(
                        return_value=mock.Mock(
                            content='<?xml version="1.0" encoding="UTF-8"?> <soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns="http://soap.sforce.com/2006/04/metadata"><soapenv:Body><checkRetrieveStatusResponse><result><done>true</done><fileProperties><createdById>0051m0000069dpxAAA</createdById><createdByName>User User</createdByName><createdDate>2023-10-09T08:47:44.875Z</createdDate><fileName>unpackaged/package.xml</fileName><fullName>unpackaged/package.xml</fullName><id></id><lastModifiedById>0051m0000069dpxAAA</lastModifiedById><lastModifiedByName>User User</lastModifiedByName><lastModifiedDate>2023-10-09T08:47:44.875Z</lastModifiedDate><manageableState>unmanaged</manageableState><type>Package</type></fileProperties><id>09S1m000001EKBkEAO</id><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery_Item__c.Food_Expiration_Date__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery__c.Status__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery_Item__c.Food_Storage__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery_Item__c.Delivery__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomField&apos; named &apos;Delivery__c.Supplier__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomObject&apos; named &apos;Delivery__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomObject&apos; named &apos;Delivery_Item__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;CustomTab&apos; named &apos;Delivery__c&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;Layout&apos; named &apos;Delivery_Item__c-Delivery Item Layout&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;Layout&apos; named &apos;Delivery__c-Delivery Layout&apos; cannot be found</problem></messages><messages><fileName>unpackaged/package.xml</fileName><problem>Entity of type &apos;ListView&apos; named &apos;Delivery__c.All&apos; cannot be found</problem></messages><status>Succeeded</status><success>true</success><zipFile>UEsDBBQACAgIAPZFSVcAAAAAAAAAAAAAAAAWAAAAdW5wYWNrYWdlZC9wYWNrYWdlLnhtbKVTTU/DMAy991dUva8pCNCE0kyIUQkJCaQNrlGami2QNFXjwvrvyT7KuFTraE7Ji5/fsy3T2cbo8Atqp2yZRhdxEoVQSluocpVGr8tsMo1mLKAvQn6KFYQ+unRptEasbglxVlSxe7e1hFhaQy6T5IYkV8QAikKgiFgQ+kOxrcDt77u3AZN7STYHrbx2yx8RDOcy/gW4pKSLGsDLrC34w6ZStUBfCJ8LhH/lWKCtfZ1DuZ62kGsoGg3FWaJbIgps3DmEpqq0grqHUgoD7L5xaE2mQBeU7JD9BMifEQwex0BjJ8w85x8gcYSbkwpLkY8vdtIB4RYIn0RrGxzYgSO3n7bz132P6EV8p3V/fuXwTcF3v8Jh1dn1NE4o6V4BJYcNZ8EPUEsHCGtr2HwiAQAAEwQAAFBLAQIUABQACAgIAPZFSVdra9h8IgEAABMEAAAWAAAAAAAAAAAAAAAAAAAAAAB1bnBhY2thZ2VkL3BhY2thZ2UueG1sUEsFBgAAAAABAAEARAAAAGYBAAAAAA==</zipFile></result></checkRetrieveStatusResponse></soapenv:Body></soapenv:Envelope>'
                        )
                    )
                )
            )

            is_collision, xml_map = task._collision_check(path)

            assert is_collision is True
            assert xml_map == {
                "CustomField": ["Delivery__c.Scheduled_Date__c"],
                "CustomObject": [],
                "CustomTab": [],
                "Layout": [],
                "ListView": [],
            }

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_get_api__managed(self, mock_org_config, rest_deploy):
        with temporary_dir() as path:
            touch("package.xml")
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "namespace_inject": "ns",
                    "unmanaged": False,
                    "rest_deploy": rest_deploy,
                },
            )

            api = task._get_api()
            zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(api.package_zip)), "r")
            assert "package.xml" in zf.namelist()
            zf.close()

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_get_api__additional_options(self, mock_org_config, rest_deploy):
        with temporary_dir() as path:
            touch("package.xml")
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "test_level": "RunSpecifiedTests",
                    "specified_tests": "TestA,TestB",
                    "unmanaged": False,
                    "rest_deploy": rest_deploy,
                },
            )

            api = task._get_api()
            assert api.run_tests == ["TestA", "TestB"]
            assert api.test_level == "RunSpecifiedTests"

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_get_api__skip_clean_meta_xml(self, mock_org_config, rest_deploy):
        with temporary_dir() as path:
            touch("package.xml")
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "clean_meta_xml": False,
                    "unmanaged": True,
                    "rest_deploy": rest_deploy,
                },
            )

            api = task._get_api()
            zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(api.package_zip)), "r")
            assert "package.xml" in zf.namelist()
            zf.close()

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_get_api__static_resources(self, mock_org_config, rest_deploy):
        with temporary_dir() as path:
            with open("package.xml", "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <types>
        <name>OtherType</name>
    </types>
</Package>"""
                )
                touch("otherfile")

            with temporary_dir() as static_resource_path:
                os.mkdir("TestBundle")
                touch("TestBundle/test.txt")
                touch("TestBundle.resource-meta.xml")

                task = create_task(
                    Deploy,
                    {
                        "path": path,
                        "static_resource_path": static_resource_path,
                        "namespace_tokenize": "ns",
                        "namespace_inject": "ns",
                        "namespace_strip": "ns",
                        "unmanaged": True,
                        "rest_deploy": rest_deploy,
                    },
                )

                api = task._get_api()
                zf = zipfile.ZipFile(io.BytesIO(base64.b64decode(api.package_zip)), "r")
                namelist = zf.namelist()
                assert "staticresources/TestBundle.resource" in namelist
                assert "staticresources/TestBundle.resource-meta.xml" in namelist
                package_xml = zf.read("package.xml").decode()
                assert "<name>StaticResource</name>" in package_xml
                assert "<members>TestBundle</members>" in package_xml
                zf.close()

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_get_api__missing_path(self, mock_org_config, rest_deploy):
        task = create_task(
            Deploy,
            {
                "path": "BOGUS",
                "unmanaged": True,
                "rest_deploy": rest_deploy,
            },
        )

        api = task._get_api()
        assert api is None

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_get_api__empty_package_zip(self, mock_org_config, rest_deploy):
        with temporary_dir() as path:
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "unmanaged": True,
                    "rest_deploy": rest_deploy,
                },
            )

            api = task._get_api()
            assert api is None

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    def test_get_api__collision_detected(self, mock_org_config):
        with temporary_dir() as path:
            task = create_task(
                Deploy,
                {
                    "path": path,
                    "namespace_tokenize": "ns",
                    "namespace_inject": "ns",
                    "namespace_strip": "ns",
                    "unmanaged": True,
                    "collision_check": True,
                },
            )
            task._get_package_zip = mock.Mock(return_value={"Custom Field": []})

            api = task._get_api()
            assert api is None

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_init_options(self, mock_org_config, rest_deploy):
        with pytest.raises(TaskOptionsError):
            create_task(
                Deploy,
                {
                    "path": "empty",
                    "test_level": "RunSpecifiedTests",
                    "unmanaged": False,
                    "rest_deploy": rest_deploy,
                },
            )

        with pytest.raises(TaskOptionsError):
            create_task(
                Deploy, {"path": "empty", "test_level": "Test", "unmanaged": False}
            )

        with pytest.raises(TaskOptionsError):
            create_task(
                Deploy,
                {
                    "path": "empty",
                    "test_level": "RunLocalTests",
                    "specified_tests": ["TestA"],
                    "unmanaged": False,
                    "rest_deploy": rest_deploy,
                },
            )

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_init_options__transforms(self, mock_org_config, rest_deploy):
        d = create_task(
            Deploy,
            {
                "path": "src",
                "transforms": ["clean_meta_xml"],
                "rest_deploy": rest_deploy,
            },
        )

        assert len(d.transforms) == 1
        assert isinstance(d.transforms[0], CleanMetaXMLTransform)

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_init_options__bad_transforms(self, mock_org_config, rest_deploy):
        with pytest.raises(TaskOptionsError) as e:
            create_task(
                Deploy,
                {
                    "path": "src",
                    "transforms": [{}],
                    "rest_deploy": rest_deploy,
                },
            )

            assert "transform spec is not valid" in str(e)

    @mock.patch(
        "cumulusci.core.config.org_config.OrgConfig.installed_packages", return_value=[]
    )
    @pytest.mark.parametrize("rest_deploy", [True, False])
    def test_freeze_sets_kind(self, mock_org_config, rest_deploy):
        task = create_task(
            Deploy,
            {
                "path": "path",
                "namespace_tokenize": "ns",
                "namespace_inject": "ns",
                "namespace_strip": "ns",
                "rest_deploy": rest_deploy,
            },
        )
        step = StepSpec(
            step_num=1,
            task_name="deploy",
            task_config=task.task_config,
            task_class=None,
            project_config=task.project_config,
        )

        assert all(s["kind"] == "metadata" for s in task.freeze(step))



from pathlib import Path
from unittest import mock
from unittest.mock import ANY

import pytest
from sarge import Capture

from cumulusci.core.config import TaskConfig
from cumulusci.core.config.project_config import BaseProjectConfig
from cumulusci.core.config.universal_config import UniversalConfig
from cumulusci.tasks.dx_convert_from import DxConvertFrom
from cumulusci.utils import temporary_dir


@pytest.fixture
def project_config():
    universal_config = UniversalConfig()
    project_config = BaseProjectConfig(universal_config, config={"no_yaml": True})
    project_config.project__name = "TestProject"
    return project_config


@pytest.fixture
def task_config():
    return TaskConfig({"options": {"src_dir": "src"}})


@pytest.fixture
def sarge_process():
    p = mock.Mock()
    p.returncode = 0
    p.stdout = Capture()
    p.stderr = Capture()
    return p


@pytest.fixture
def dx_convert_task(project_config, task_config):
    return DxConvertFrom(project_config, task_config)


@mock.patch("cumulusci.tasks.command.sarge")
def test_dx_convert_from(sarge, sarge_process, dx_convert_task):
    """Ensure that we clear out the `src/`"""
    with temporary_dir():
        src_dir = Path("src")
        src_dir.mkdir(exist_ok=True)

        sarge.Command.return_value = sarge_process
        dx_convert_task()

        assert not src_dir.exists()
        sarge.Command.assert_called_once_with(
            "sf project convert source -d src",
            cwd=".",
            env=ANY,
            shell=True,
            stdout=ANY,
            stderr=ANY,
        )
